/*  Copyright (c) 2002, 2003, 2006 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  David M. Richter <richterd@citi.umich.edu>
 *  Andy Adamson <andros@citi.umich.edu>
 *  Alexis Mackenzie <allamack@citi.umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <attr/xattr.h>
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <libgen.h>
#include "libacl_nfs4.h"

/* Actions */
#define NO_ACTION		0
#define MODIFY_ACTION		1
#define SUBSTITUTE_ACTION	2
#define REMOVE_ACTION		3
#define INSERT_ACTION		4
#define EDIT_ACTION		5

/* Walks */
#define DEFAULT_WALK		0	/* Follow symbolic link args, Skip links in subdirectories */
#define LOGICAL_WALK		1	/* Follow all symbolic links */
#define PHYSICAL_WALK		2	/* Skip all symbolic links */

/* Recursion */
#define NO_RECURSIVE		0
#define YES_RECURSIVE		1

#define MKTMPLATE  		"/tmp/.nfs4_setfacl-tmp-XXXXXX"
#define EDITOR  		"vi"  /* <- evangelism! */
#define u32 u_int32_t

#define is_flag(s, t)	 (!(strcmp(s, t)))
#define has_arg(i, a)	 ((i + 2) < a)
#define assert_wu_wei(a)	 \
	do { \
		if (a != NO_ACTION) { \
			fprintf(stderr, "More than one action specified.\n"); \
			usage(); \
			goto failed; \
		} \
	} while (0)

void more_usage();
void __usage(const char *, int);
static int is_editfacl = 0;
#define usage()	__usage(basename(argv[0]), is_editfacl)


/* XXX: things we need to handle:
 *
 *  - we need some sort of 'purge' operation that completely clears an ACL.
 *  - qc like setfacl's default-only/custom-only modes
 *  - recursion w/ walks
 */

int main(int argc, char** argv)
{
	int res, i, is_test = 0, index = -1;
	int numpaths = 0, curpath = 0, tmp_fd = -1;
	int do_recursive = NO_RECURSIVE;
	int walk_type = DEFAULT_WALK;
	int action = NO_ACTION;
	char *arg, *path = NULL, **paths, *spec_file = NULL, *mod_string = NULL;
	char *editor = NULL, *from_ace = NULL, *to_ace = NULL;
	char tmp_name[strlen(MKTMPLATE) + 1];
	FILE *tmp_fp = NULL, *s_fp = NULL;
	struct nfs4_acl *acl;
	struct stat st;

	if (!strcmp(basename(argv[0]), "nfs4_editfacl")) {
		action = EDIT_ACTION;
		is_editfacl = 1;
	}

	if (argc == 1) {
		usage();
		goto failed;
	}

	for (i = 1; i < argc; i++) {
		arg = argv[i];

		/* too bad getopt doesn't lend itself to, e.g., -x  acl_spec | index | -i index  */
		if ((is_flag(arg, "-a") || is_flag(arg, "-A"))) {
			assert_wu_wei(action);
			action = INSERT_ACTION;

			if (!has_arg(i, argc)) {
				fprintf(stderr, "Sorry, -a requires an 'acl_spec', whilst -A requires a 'spec_file'.\n");
				goto failed;
			}

			if (is_flag(arg, "-a"))
				mod_string = argv[++i];
			else
				spec_file = argv[++i];
		
			/* run along if no more args (defaults to index 1 == prepend)*/
			if (!has_arg(i, argc))
				continue;
			
			arg = argv[++i];
			index = strtoul_reals(arg, 10);
			if (index == ULONG_MAX) {
				/* oops, it wasn't an index; reset */
				i--;
				index = -1;
			} else if (index == 0) {
				fprintf(stderr, "Sorry, valid indices start at '1'.\n");
				goto failed;
			}
		} 
		else if (is_flag(arg, "-s")) {
			assert_wu_wei(action);
			action = SUBSTITUTE_ACTION;

			if (!has_arg(i, argc)) {
				fprintf(stderr, "Sorry, -s requires an 'acl_spec'.\n");
				goto failed;
			}
			mod_string = argv[++i];
		}
		else if (is_flag(arg, "-S")) {
			assert_wu_wei(action);
			action = SUBSTITUTE_ACTION;

			if (!has_arg(i, argc)) {
				fprintf(stderr, "Sorry, -S requires a 'spec_file'.\n");
				goto failed;
			}
			spec_file = argv[++i];
		}
		else if (is_flag(arg, "-x")) {
			assert_wu_wei(action);
			action = REMOVE_ACTION;

			if (!has_arg(i, argc)) {
				fprintf(stderr, "Sorry, -x requires either an 'acl_spec' or the specific index of an entry to remove.\n");
				goto failed;
			}
			arg = argv[++i];
			index = strtoul_reals(arg, 10);

			/* if it wasn't an index number and doesn't look like a flag, is there an acl_spec? */
			if (index == ULONG_MAX) {
				/* oops, it's a flag; continue */
				if ('-' == *arg)
					i--;
				else
					mod_string = arg;
			} else if (index == 0) {
				fprintf(stderr, "Sorry, valid indices start at '1'.\n");
				goto failed;
			}
		}
		else if (is_flag(arg, "-X")) {
			assert_wu_wei(action);
			action = REMOVE_ACTION;

			if (!has_arg(i, argc)) {
				fprintf(stderr, "Sorry, -X requires a 'spec_file'.\n");
				goto failed;
			} 
			spec_file = argv[++i];
		}
		else if (is_flag(arg, "-m")) {
			assert_wu_wei(action);
			action = MODIFY_ACTION;

			if (!has_arg(i + 1, argc)) {
				fprintf(stderr, "Sorry, -m requires 'from_ace' and 'to_ace' arguments.\n");
				goto failed;
			}
			from_ace = argv[++i];
			to_ace = argv[++i];
		}
		/* XXX: maybe we should go back to getopt. */
		else if (is_flag(arg, "-e") || is_flag(arg, "--edit")) {
			assert_wu_wei(action);
			action = EDIT_ACTION;
		}
		else if (is_flag(arg, "--test")) {
			is_test = 1;
		}
		else if (is_flag(arg, "-h") || is_flag(arg, "-?") || is_flag(arg, "--help")) {
			usage();
			return 0;
		}
		else if (is_flag(arg, "--version")) {
			printf("%s %s\n", basename(argv[0]), VERSION);
			return 0;
		} 
		else if (is_flag(arg, "--more-help")) {
			more_usage();
			printf("\nSorry, this isn't done ...\n");
			return 0;
		}
		else {
			paths = &argv[i];
			numpaths = argc - i;
			break;
		}
//			case 'P':
//				if (walk_type != DEFAULT_WALK) {
//					fprintf(stderr, "More than one walk type specified\n");
//					usage();
//					goto failed;
//				}
//				walk_type = PHYSICAL_WALK;
//				break;
//
//			case 'R':
//				do_recursive = YES_RECURSIVE;
//				break;
//			case 'L':
//				if (walk_type != DEFAULT_WALK) {
//					fprintf(stderr, "More than one Walk type specified\n");
//					usage();
//					goto failed;
//				}
//				walk_type = LOGICAL_WALK;
//				break;
	}

	if (action == NO_ACTION) {
		fprintf(stderr, "No action specified.\n");
		goto failed;
	} else if (numpaths < 1) {
		fprintf(stderr, "No path(s) specified.\n");
		goto failed;
	} 

	if (action == EDIT_ACTION)
		if ((editor = getenv("EDITOR")) == NULL)
			editor = EDITOR;
	if (spec_file) {
		if (!strcmp(spec_file, "-")) {
			s_fp = stdin;
		} else {
			s_fp = fopen(spec_file, "r");
			if (s_fp == NULL) {
				fprintf(stderr, "Access error on spec file: %s\n", spec_file);
				goto failed;
			}
		}
		if ((mod_string = nfs4_acl_spec_from_file(s_fp)) == NULL) {
			if (s_fp == stdin)
				spec_file = "(stdin)";
			fprintf(stderr, "Failed to create ACL from contents of 'spec_file' %s.\n", spec_file);
			goto failed;
		}
	}

	while (numpaths > curpath) {
		path = paths[curpath++];
		res = stat(path, &st);
		if (res < 0) {
			perror(path);
			goto failed;
		}

		if (action == SUBSTITUTE_ACTION)
			acl = nfs4_new_acl(st.st_mode & S_IFDIR ? 1 : 0);
		else
			acl = nfs4_acl_for_path(path);

		if (acl == NULL) {
			fprintf(stderr, "Failed to instantiate ACL.\n");
			goto failed;
		}
		
		if (action == INSERT_ACTION) {
			/* default to prepending */
			if (index < 1)
				index = 1;

			if (nfs4_insert_string_aces(acl, mod_string, (index - 1))) {
				fprintf(stderr, "Failed while inserting ACE(s) (at index %d).\n", index);
				goto failed_free;
			}
		} else if (action == REMOVE_ACTION) {
			if (index != -1) {
				/* "(index - 1)" because we access the ACL zero-based-wise, but the CLI arg is one-based. */
				if ((index - 1) > acl->naces) {
					fprintf(stderr, "Index %u is out of range (%d ACEs in ACL)\n", index, acl->naces);
					goto failed_free;
				}
				if (nfs4_remove_ace_at(acl, (index - 1))) {
					fprintf(stderr, "Failed to remove ACE at index %u.\n", index);
					goto failed_free;
				}
			} else if (nfs4_remove_string_aces(acl, mod_string)) {
				fprintf(stderr, "Failed while removing matched ACE(s).\n");
				goto failed_free;
			}
		} else if (action == MODIFY_ACTION) {
			if (nfs4_replace_ace_spec(acl, from_ace, to_ace)) {
				fprintf(stderr, "Failed while trying to replace ACE.\n");
				goto failed_free;
			}
		} else if (action == SUBSTITUTE_ACTION) {
			if (nfs4_insert_string_aces(acl, mod_string, 0)) {
				fprintf(stderr, "Failed while inserting ACE(s).\n");
				goto failed_free;
			}
		} else if (action == EDIT_ACTION) {
			char *edargv[3];
			int edpid, w, status;

			strcpy(tmp_name, MKTMPLATE);
			if ((tmp_fd = mkstemp(tmp_name)) == -1) {
				fprintf(stderr, "Unable to make tempfile \"%s\" for editing.\n", tmp_name);
				goto failed_free;
			}
			if ((tmp_fp = fdopen(tmp_fd, "w+")) == NULL) {
				fprintf(stderr, "Unable to fdopen tempfile \"%s\" for editing.\n", tmp_name);
				goto failed_free;
			}
			nfs4_print_acl(tmp_fp, acl);
			nfs4_free_acl(acl);
		
			edargv[0] = editor;
			edargv[1] = tmp_name;
			edargv[2] = (char *) NULL;

			edpid = fork();
			if (edpid == 0) {
				// kid, dup/close in/out/err
				execvp(edargv[0], edargv);
				// kid, if here, error, exit kablooie
				fprintf(stderr, "Failed to exec() editor \"%s\".\n", editor);
				goto failed;
			} else if (edpid == -1 ) {
				fprintf(stderr, "FAILED TO FORK EDITOR\n");
				goto failed;
			}

			// rent, maybe block some signals
			// rent, maybe loop around: while (w = waitpid(kpid) == ERR && errno == EINTR) 
			if ((w = waitpid(edpid, &status, 0)) == -1) {
				fprintf(stderr, "FAILED WAITPIDING\n");
			} else if (WIFEXITED(status)) {
				res = WEXITSTATUS(status);
			} else if (WIFSIGNALED(status)) {
				fprintf(stderr, "SIGNALED OUT OF EDITING\n");
				res = -1;
			}

			if (res != 0) {
				fprintf(stderr, "Editor `%s' did not exit cleanly; changes will not be saved.\n", editor);
				goto failed;
			}
			if ((acl = nfs4_new_acl(st.st_mode & S_IFDIR ? 1 : 0)) == NULL) {
				fprintf(stderr, "Failed creating new ACL from tempfile %s. Aborting.\n", tmp_name);
				goto failed;
			}
			rewind(tmp_fp);
			if (nfs4_insert_file_aces(acl, tmp_fp, 0)) {
				fprintf(stderr, "Failed loading ACL from edit tempfile %s. Aborting.\n", tmp_name);
				goto failed_free;
			}

			unlink(tmp_name);
			tmp_fd = -1;
		}
	
		if (is_test) {
			fprintf(stderr, "## Test mode only - the resulting ACL for \"%s\": \n", path);
			nfs4_print_acl(stdout, acl);
		} else
			res = nfs4_set_acl(acl, path);
	
		nfs4_free_acl(acl);
	}
out:
	if (tmp_fd != -1)
		unlink(tmp_name);
	return res;

failed_free:
	nfs4_free_acl(acl);
failed:
	res = 1;
	goto out;
}


void __usage(const char *name, int is_ef)
{
	static const char *sfoosage = \
	"%s %s -- manipulate NFSv4 file/directory access control lists\n"
	"Usage: %s [OPTIONS] COMMAND file ...\n"
	" .. where COMMAND is one of:\n"
	"   -a acl_spec [index]	 add ACL entries in acl_spec at index (DEFAULT: 1)\n"
	"   -A file [index]	 read ACL entries to add from file\n"
	"   -x acl_spec | index	 remove ACL entries or entry-at-index from ACL\n"
	"   -X file  		 read ACL entries to remove from file\n"
	"   -s acl_spec		 set ACL to acl_spec (replaces existing ACL)\n"
	"   -S file		 read ACL entries to set from file\n"
	"   -e, --edit 		 edit ACL in $EDITOR (DEFAULT: " EDITOR "); save on clean exit\n"
	"   -m from_ace to_ace	 modify in-place: replace 'from_ace' with 'to_ace'\n"
	"   --version		 print version and exit\n"
	"   -?, -h, --help 	 display this text and exit\n"
	"   --more-help		 DEPRECATED; display terms and examples and exit\n"
	"\n"
	"     NOTE: if \"-\" is given with -A/-X/-S, entries will be read from stdin.\n"
	"\n"
	" .. and where OPTIONS is any (or none) of:\n"
	"   --test	 	 print resulting ACL, do not save changes\n"
	"   -R (unimplemented) 	 recursively apply to all files and directories\n";

	static const char *efoosage = \
	"%s %s -- edit NFSv4 file/directory access control lists\n"
	"Usage: %s [OPTIONS] file ...\n"
	" .. where OPTIONS is any (or none) of:\n"
	"   --test	 	 print resulting ACL, do not save changes\n"
	"   -R (unimplemented) 	 recursively apply to all files and directories\n";

	const char *oosage = is_ef ? efoosage : sfoosage;

	fprintf(stderr, oosage, name, VERSION, name);
}

void more_usage()
{
	const char *more_oosage = \
	"-----------\n"
	" TERMS\n"
	"-----------\n"
	"* \"ace_spec\" - a single NFSv4 ACE, which is a colon-delimited, 4-field string:\n"
	"\n"
	"    <type>:<flags>:<principal>:<access>\n"
	"\n"
	"    * <type> - one of: 'A' (allow), 'D' (deny), 'U' (audit), 'L' (alarm)\n"
	"    * <principal> - name of user or group, or one of: \"OWNER@\", \"GROUP@\", \"EVERYONE@\"\n"
	"    * <flags> - one or more of:\n"
	"        'f' file inherit ACE\n"
	"        'd' directory inherit ACE\n"
	"        'p' no-propagate inherit\n"
	"        'i' inherit-only\n"
	"        'S' successful access\n"
	"        'F' failed access\n"
	"\n"
	"    * <access> - one or more of:\n"
	"        'r' read data / list directory \n"
	"        'w' write data / create file \n"
	"        'a' append data / create subdirectory \n"
	"        'x' execute \n"
	"        'd' delete\n"
	"        'D' delete child (directories only)\n"
	"        't' read attrs\n"
	"        'T' write attrs\n"
	"        'n' read named attrs\n"
	"        'N' write named attrs\n"
	"        'c' read ACL\n"
	"        'C' write ACL\n"
	"        'o' write owner\n"
	"        'y' synchronize\n"
	"\n"
	"* \"acl_spec\" - one or more ace_spec(s), separated by commas or whitespace.\n"
	"\n"
	"* \"spec_file\" - a file containing an acl_spec.  Blank lines are skipped, and \n"
	"    any text following a pound-sign ('#') and up to a newline is treated as\n"
	"    a comment.\n"
	"\n"
	"-----------\n"
	" EXAMPLES\n"
	"-----------\n"
	"* sample contents of a spec_file:\n"
	"	A::OWNER@:rwanNtTcCy\n"
	"	D::OWNER@:x\n"
	"	A:g:GROUP@:rntcy\n"
	"	D:g:GROUP@:waxTC\n"
	"	A::EVERYONE@:rntcy\n"
	"	D::EVERYONE@:waxTC\n"
	"\n"
	"* add ACE \"A::alice@nfsdomain.org:rntcy\" (defaults to prepending):\n"
	"	$ nfs4_setfacl -a A::alice@nfsdomain.org:rntcy afile\n"
	"\n"
	"* set ACL (overwrites original) to contents of spec_file \"sample_spec_file\":\n"
	"	$ nfs4_setfacl -S sample_spec_file afile\n"
	"\n"
	"* edit existing ACL in a text editor and set ACL on clean exit:\n"
	"	$ nfs4_setfacl -e afile\n"
	"\n"
	"* delete the ACEs \"A::EVERYONE@:rntcy\" and \"D::EVERYONE@:waxTC\":\n"
	"	$ nfs4_setfacl -x \"A::EVERYONE@:rntcy, D::EVERYONE@:waxTC\" afile\n"
	"\n"
	"* delete the first ACE, but only print the resulting ACL:\n"
	"	$ nfs4_setfacl --test -x 1 afile\n"
	"\n"
	"* delete ACE(s) found in the spec_file \"sample_spec_file\":\n"
	"	$ nfs4_setfacl -X sample_spec_file afile\n"
	"\n"
	"* change ACE \"D::OWNER@:x\" to \"D::OWNER@:xNTCoy\":\n"
	"	$ nfs4_setfacl -m D::OWNER@:x  D::OWNER@:xNTCoy afile\n"
	"	\n"
	"* set ACLs of \"bar\" and \"frobaz\" to ACL of \"foo\":\n"
	"	$ nfs4_getfacl foo | nfs4_setfacl -S - bar frobaz\n";

	printf("%s", more_oosage); 
}

#if 0
void usage(void)
{
	fprintf(stderr, "usage: nfs4_setfacl [-jRLP] [{-m|-x} acl_spec] [{-M|-X} acl_file] file ...\n");
	fprintf(stderr, "  -S spec_file	 set NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -M spec_file	 modify NFSv4 ACL with contents of spec_file\n");
	fprintf(stderr, "  -m acl_spec	 modify NFSv4 ACL with contents of acl_spec\n");
	fprintf(stderr, "  -s acl_spec	 set NFSv4 ACL with acl_spec string\n");
	fprintf(stderr, "  -a acl_spec   append NFSv4 ACL entry onto existing acl\n");
	fprintf(stderr, "  -A spec_file	 append contents of spec_file onto existing NFSv4 ACL\n");
	fprintf(stderr, "  -x acl_spec	 remove entries from the ACL(s) of file(s)\n");
	fprintf(stderr, "  -X spe_file	 read ACL entries to remove from spec file\n");
	fprintf(stderr, "  -j		 format ACL to LINUX POSIX translation spec\n");
	fprintf(stderr, "  -R		 recurse into subdirectories\n");
	fprintf(stderr, "  -L		 logical walk, follow symbolic links\n");
	fprintf(stderr, "  -P		 physical walk, do not follow symbolic links\n");
	return;
}
#endif
